"""Thanks DK: Source, https://gist.github.com/ddkasa/0a60dcb1bd13e67bc79cc6f3b144aa8d"""

from __future__ import annotations

from operator import itemgetter
from typing import TYPE_CHECKING

from textual.box_model import BoxModel
from textual.css.styles import RenderStyles
from textual._resolve import resolve_box_models
from textual.geometry import (
    NULL_OFFSET,
    NULL_SIZE,
    Region,
    Size,
    Spacing,
)
from textual.layout import Layout, WidgetPlacement
from textual.widget import Widget

from math import floor

if TYPE_CHECKING:
    from textual.layout import ArrangeResult


class FlexBoxLayout(Layout):
    name = "flexbox"

    def arrange(
        self,
        parent: Widget,
        children: list[Widget],
        size: Size,
    ) -> ArrangeResult:
        """Generate a layout map that defines where the widgets will be drawn.
        Args:
            parent: Parent widget.
            children: A list of widgets to be placed.
            size: Size of container.
        Returns:
            An iterable of widget locations.
        """
        parent.pre_layout(self)
        viewport = parent.app.size

        styles = [child.styles for child in children]
        placements = [
            (
                style.get_rule("position") == "absolute",
                style.overlay == "screen",
            )
            for style in styles
        ]
        margins: list[Spacing] = [
            style.margin for style, (overlay, _) in zip(styles, placements) if overlay
        ]

        NewSize = Size
        if margins:
            resolve_margin = NewSize(
                sum(
                    [
                        max(margin1[1], margin2[3])
                        for margin1, margin2 in zip(margins, margins[1:])
                    ]
                )
                + (margins[0].left + margins[-1].right),
                max(
                    [
                        margin_top + margin_bottom
                        for margin_top, _, margin_bottom, _ in margins
                    ]
                ),
            )
        else:
            resolve_margin = NULL_SIZE

        regions = self._resolve_regions(
            parent,
            children,
            viewport,
            size,
            resolve_margin,
            placements,
            styles,
        )

        NewPlacement = WidgetPlacement

        return [
            NewPlacement(
                region=region,
                offset=style.offset.resolve(
                    NewSize(region.width, region.height),
                    viewport,
                )
                if style.has_rule("offset")
                else NULL_OFFSET,
                margin=box_model.margin,
                widget=widget,
                order=i,
                overlay=is_overlay,
                absolute=is_abs,
            )
            for i, (
                widget,
                (region, box_model),
                (is_abs, is_overlay),
                style,
            ) in enumerate(zip(children, regions, placements, styles, strict=False))
        ]

    def _resolve_regions(
        self,
        parent: Widget,
        children: list[Widget],
        viewport: Size,
        size: Size,
        margin: Size,
        placements: list[tuple[bool, bool]],
        styles: list[RenderStyles],
    ) -> list[tuple[Region, BoxModel]]:
        """Generate regions balanced rows from provided widgets.
        Args:
            parent: The parent container widget primarily for styles.
            children: Widgets to be layed out.
            viewport: Size of the of parent viewport.
            size: Given size of the container.
            margin: Initial size of margin for resolving box models.
            placements: Booleans for screening and absolute placement setting.
            styles: Styles of children expanded out beforehand.
        Returns:
            Tuples of regions and box models ready to be placed.
        """
        margin_width, margin_height = margin
        pwidth, pheight = viewport - margin
        sizes = [(style.width, style.height) for style in styles]

        parent_width = size.width

        resolved = resolve_box_models(
            list(map(itemgetter(0), sizes)),
            children,
            size,
            viewport,
            margin,
            resolve_dimension="width",
        )

        row_sizes: list[tuple[int | None, int, int, int]] = []
        """Tuples of row indexes. 1. Index 2. row_width 3. row-y-pos, 4. row_height"""
        add_row = row_sizes.append
        row_width, row_pos, max_height = 0, 0, 0
        for i, (
            widget,
            (width, height, box_margin),
            (is_abs, is_overlay),
        ) in enumerate(zip(children, resolved, placements, strict=True)):
            if is_abs or is_overlay:
                continue
            box_width = floor(width + box_margin.width)
            next_width = floor(row_width + box_width)
            if next_width > parent_width:
                add_row((i, row_width, row_pos, max_height))
                next_width = box_width
                row_pos += max_height
                max_height = 0
            next_height = floor(height + box_margin.height)
            max_height = max_height if max_height >= next_height else next_height
            row_width = next_width

        add_row((None, row_width, row_pos, max_height))
        halign, valign = parent.styles.align
        center_aligned = halign == "center"
        right_aligned = halign == "right"
        NewRegion = Region
        box_models: list[Region] = []
        add_region = box_models.append
        prev_index = None
        for i in range(len(row_sizes)):
            index, row_width, row_pos, max_height = row_sizes[i]
            if center_aligned:
                x = floor((parent_width - row_width) / 2)
            elif right_aligned:
                x = floor(parent_width - row_width)
            else:
                x = 0

            for widget, (width, height, box_margin), (is_abs, is_overlay) in zip(
                children[prev_index:index],
                resolved[prev_index:index],
                placements[prev_index:index],
                strict=True,
            ):
                add_region(
                    NewRegion(
                        x + box_margin.left,
                        row_pos + box_margin.top,
                        width := floor(width),
                        floor(height),
                    )
                )
                if not is_abs and not is_overlay:
                    x += width + box_margin.width

            prev_index = index

        return list(zip(box_models, resolved))


class FlexBoxContainer(Widget):
    """Container widget with predefined flexbox layout."""

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        markup: bool = True,
    ) -> None:
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            markup=markup,
        )

        self._layout = FlexBoxLayout()

    @property
    def layout(self) -> FlexBoxLayout:
        return self._layout



import pytest
from textual.app import App

from textual_tags import Tags

TEST_TAGS = ["Tag1", "Tag2", "Tag3"]


class TestApp(App):
    def compose(self):
        self.tags_widget = Tags(tag_values=TEST_TAGS)
        yield self.tags_widget


@pytest.fixture
def test_app():
    return TestApp()



from __future__ import annotations
from typing import Iterable

from textual.binding import Binding
from textual.reactive import reactive
from textual.app import RenderResult
from textual.events import Key
from textual.message import Message
from textual.widgets import Input, Label
from textual_autocomplete import (
    AutoComplete,
    DropdownItem,
    TargetState,
    AutoCompleteList,
)

from textual_tags.flexbox import FlexBoxContainer


class TagAutoComplete(AutoComplete):
    class Applied(Message):
        def __init__(self, autocomplete: TagAutoComplete) -> None:
            self.autocomplete = autocomplete
            super().__init__()

        @property
        def control(self):
            return self.autocomplete

    def post_completion(self):
        self.action_hide()
        self.post_message(self.Applied(autocomplete=self))


class TagInput(Input):
    DEFAULT_CSS = """
    TagInput {
        margin:0;
        padding:0;
        height: auto;
        width: auto;
        min-width:20;
        border:none;

        & > :focus {
            border:none;
        }
    }
    """

    def on_focus(self):
        self.parent.query_one(TagAutoComplete).action_show()

    async def on_key(self, event: Key):
        if event.key == "backspace":
            if not self.value:
                await self.parent.reset_last_tag()


class Tag(Label):
    DEFAULT_CSS = """
    Tag {
        margin:0 1 0 0 ;
        color:$primary-lighten-2;
        background:$panel;

        &:hover {
            background:$primary-darken-2;
            tint: 10%;
        }
        &:focus {
            background:$primary-darken-2;
            tint: 10%;
        }

    }
    """
    RIGHT_END = "\ue0b4"
    LEFT_END = "\ue0b6"

    show_x: reactive[bool] = reactive(False)
    can_focus = True

    class Removed(Message):
        def __init__(self, tag: Tag) -> None:
            self.tag = tag
            super().__init__()

        @property
        def control(self):
            return self.tag

    class Focused(Message):
        def __init__(self, tag: Tag) -> None:
            self.tag = tag
            super().__init__()

        @property
        def control(self):
            return self.tag

    class Hovered(Message):
        def __init__(self, tag: Tag) -> None:
            self.tag = tag
            super().__init__()

        @property
        def control(self):
            return self.tag

    def render(self) -> RenderResult:
        background = self.styles.background.hex
        parent_background = self.colors[0].hex
        # Add extra padding with single space if self.show_x == False
        # To prevent layout change if its toggled
        left_round_part = f"[{background} on {parent_background}]{self.LEFT_END}[/]"
        right_round_part = f"[{background} on {parent_background}]{self.RIGHT_END}[/]"
        label_part = f"{self.value}" if self.show_x else f" {self.value}"
        x_part = " x" if self.show_x else " "
        return left_round_part + label_part + x_part + right_round_part

    def on_click(self):
        self.remove()

    def on_key(self, event: Key):
        if event.key == "enter":
            self.remove()
            event.prevent_default()
            event.stop()

    def watch_show_x(self):
        self.render()

    def on_prune(self):
        self.post_message(self.Removed(self))

    def on_focus(self):
        self.post_message(self.Focused(self))

    def on_enter(self):
        self.post_message(self.Hovered(self))

    @property
    def value(self):
        return self.renderable


class Tags(FlexBoxContainer):
    DEFAULT_CSS = """
    Tags {
        padding:0 0 0 1 ;
        height:auto;
        min-height:3;
    }
    """

    BINDINGS = [
        Binding("ctrl+j", "next_hightlight", priority=True),
        Binding("ctrl+k", "previous_hightlight", priority=True),
        Binding("ctrl+o", "clear_tags", priority=True),
    ]

    tag_values: reactive[set[str]] = reactive(set())
    show_x: reactive[bool] = reactive(False)
    allow_new_tags: reactive[bool] = reactive(False)
    selected_tags: reactive[set[str]] = reactive(set())

    def __init__(
        self,
        tag_values: list | set,
        show_x: bool = False,
        start_with_tags_selected: bool = True,
        allow_new_tags: bool = False,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """A tags widget to select/unselect predefined Tags or add new ones.

        Args:
            tag_values: The available tags for this widget, will be turned into a set internally, if a list is provided
            show_x: Puts a `X` behind the actual tag-label (default=False)
            start_with_tags_selected: All Tags will be selected on initialisation (default=True)
            allow_new_tags: Allow adding any value as tag, not just predefined ones (default=False)
            id: The DOM node id of the widget.
            classes: The CSS classes of the widget.
            disabled: Whether the widget is disabled.
        """
        # Use a set to remove duplicate tags
        if isinstance(tag_values, list):
            tag_values_set = set(tag_values)

        super().__init__(id=id, classes=classes, disabled=disabled)
        self.tag_values = tag_values_set
        self.show_x = show_x
        self.allow_new_tags = allow_new_tags
        self.start_with_tags_selected = start_with_tags_selected

    async def on_mount(self):
        self.query_one(TagInput).placeholder = "Enter a tag..."
        if self.start_with_tags_selected:
            await self._populate_with_tags()

    def compose(self):
        tag_input = TagInput(id="input_tag")
        yield tag_input

        yield TagAutoComplete(
            target=tag_input,
            candidates=self.update_autocomplete_candidates,
        )

    def _on_tag_removed(self, event: Tag.Removed):
        self.selected_tags.remove(event.tag.value)
        self.mutate_reactive(Tags.selected_tags)

    def update_autocomplete_candidates(self, state: TargetState) -> list[DropdownItem]:
        return [DropdownItem(unselected_tag) for unselected_tag in self.unselected_tags]

    async def _on_tag_auto_complete_applied(self, event: TagAutoComplete.Applied):
        await event.autocomplete.target.action_submit()

    async def on_input_submitted(self, event: Input.Submitted):
        value = event.input.value.strip()
        # Prevent empty Tags
        if not value:
            return
        # Prevent non pre-defined entries if not allowed
        if (value not in self.tag_values) and (not self.allow_new_tags):
            # self.notify(
            #     title='Invalid Tag',
            #     message='Adding new tags is not allowed',
            #     severity='warning',
            #     timeout=2
            # )
            return

        # Prevent already selected tags
        if value in self.selected_tags:
            # self.notify(
            #     title='Invalid Tag',
            #     message='Tag already present',
            #     severity='warning',
            #     timeout=2
            # )
            return

        await self.add_new_tag(value=value)
        self.query_one(Input).clear()

    async def _populate_with_tags(self):
        for tag in self.tag_values:
            if tag not in self.selected_tags:
                await self.add_new_tag(value=tag)

    async def add_new_tag(self, value: str):
        """Adds a new Tag and updates self.tag_values if tag is not present"""
        await self.mount(Tag(value).data_bind(Tags.show_x), before="#input_tag")
        self.selected_tags.add(value)

        if value not in self.tag_values:
            self.tag_values.add(value)

        self.mutate_reactive(Tags.tag_values)
        self.mutate_reactive(Tags.selected_tags)

    async def action_clear_tags(self):
        await self.unselect_all_tags()
        self._put_focus_back_on_input()

    async def unselect_all_tags(self):
        """Removes all Tags"""
        await self.query(Tag).remove()

    def _put_focus_back_on_input(self):
        """Focus back on input after tag clear"""
        self.query_one(TagInput).cursor_position = 0
        self.query_one(TagInput).focus()

    def add_tag_values(self, new_values: str | Iterable[str]):
        """Add new tags to self.tag_values which holds all available Tags for this widget"""
        if isinstance(new_values, str):
            self.tag_values.add(new_values)
        else:
            self.tag_values.update(new_values)
        self.mutate_reactive(Tags.tag_values)

    async def reset_last_tag(self):
        if not self.allow_new_tags:
            return
        last_tag = self.query(Tag).last()
        self.tag_values.remove(last_tag.value)
        await last_tag.remove()
        self.query_one(TagInput).value = last_tag.value
        self.query_one(TagInput).cursor_position = len(last_tag.value)

    def action_next_hightlight(self):
        """go to next hightlight in completion option list"""
        option_list = self.query_one(TagAutoComplete).option_list
        displayed = self.query_one(TagAutoComplete).display
        highlighted = option_list.highlighted

        if displayed:
            highlighted = (highlighted + 1) % option_list.option_count
        else:
            self.query_one(TagAutoComplete).display = True
            highlighted = 0

        option_list.highlighted = highlighted

    def action_previous_hightlight(self):
        """go to previous hightlight in completion option list"""
        option_list = self.query_one(AutoCompleteList)
        displayed = self.query_one(TagAutoComplete).display
        highlighted = option_list.highlighted

        if displayed:
            highlighted = (highlighted - 1) % option_list.option_count
        else:
            self.query_one(TagAutoComplete).display = True
            highlighted = 0

        option_list.highlighted = highlighted

    async def watch_selected_tags(self):
        """hide input if all tags are selected and no new tags are allowed"""
        if self.allow_new_tags:
            return
        if not self.unselected_tags:
            self.query_one(TagInput).styles.display = "none"
        else:
            self.query_one(TagInput).styles.display = "block"

    async def watch_allow_new_tags(self):
        if self.allow_new_tags:
            self.query_one(TagInput).styles.display = "block"
        else:
            await self.watch_selected_tags()

    async def watch_tag_values(self):
        await self.watch_allow_new_tags()

    @property
    def unselected_tags(self) -> set[str]:
        return self.tag_values.difference(self.selected_tags)



from textual import on
from textual.app import App, ComposeResult
from textual.content import Content
from textual.widgets import Input, Switch, Label, Rule, Button
from textual_tags import Tags, Tag

DEMO_TAGS = [
    "uv",
    "Terminal",
    "tcss",
    "Textual",
    "Tags",
    "Widget",
    "Python",
    "TUI",
    "Highlander",
]


class DemoApp(App):
    CSS_PATH = "assets/demo.tcss"

    def compose(self) -> ComposeResult:
        yield Tags(
            tag_values=DEMO_TAGS,
            show_x=False,
            start_with_tags_selected=True,
            allow_new_tags=False,
        )
        input = Input(
            placeholder="Add more tags to internal widget list",
            id="input_adder",
            classes="demo-widgets",
        )
        input.border_title = "Add more Tags here"
        switch_x = Switch(id="switch_x", classes="demo-widgets")
        switch_x.border_title = "Show X at end of each tag [$warning](default=False)[/]"
        switch_new = Switch(id="switch_new", classes="demo-widgets")
        switch_new.border_title = "Allow New Tags [$warning](default=False)[/]"
        button_all = Button("Mount All", id="button_all", variant="success")

        yield Rule(classes="description")
        yield Label(
            Content.from_markup(
                "You can display and navigate completion options with [$success]ctrl+j/k[/] or [$success]up/down[/]"
            ),
            classes="description",
        )

        yield Rule(classes="description")
        yield Label(
            Content.from_markup(
                "Click or press enter while a tag is focused to unselect Tag, [$success]ctrl+o[/] on the tags widget unselects all"
            ),
            classes="description",
        )
        yield Rule(classes="description")
        yield input
        yield switch_x
        yield switch_new
        yield button_all

        return super().compose()

    @on(Input.Submitted, "#input_adder")
    def add_new_tag_to_widget(self, event: Input.Submitted):
        self.query_one(Tags).add_tag_values(event.input.value)
        event.input.clear()

    def on_switch_changed(self, event: Switch.Changed):
        match event.switch.id:
            case "switch_x":
                self.query_one(Tags).show_x = event.switch.value
            case "switch_new":
                self.query_one(Tags).allow_new_tags = event.switch.value

    def on_button_pressed(self, event: Button.Pressed):
        self.notify("All Tags mounted")
        tags = self.query_one(Tags)
        tags.call_later(tags._populate_with_tags)

    def on_tag_focused(self, event: Tag.Focused):
        self.notify(f"Tag {event.tag.value} focused", timeout=1)

    def on_tag_hovered(self, event: Tag.Hovered):
        self.notify(f"Tag {event.tag.value} hovered", timeout=1)


def run_demo():
    app = DemoApp()
    app.run()
